@@ -33,11 +33,15 @@ import time
 import boto3
 from git import Repo
 
+from dcicutils.beanstalk_utils import compute_ff_prd_env, compute_cgap_prd_env
 from dcicutils.env_utils import (
-    get_standard_mirror_env, data_set_for_env, get_bucket_env, INDEXER_ENVS, is_fourfront_env, is_cgap_env,
-    FF_ENV_INDEXER, CGAP_ENV_INDEXER, is_indexer_env, indexer_env_for_env
+    get_standard_mirror_env, data_set_for_env, get_bucket_env, INDEXER_ENVS,
+    is_fourfront_env, is_cgap_env, is_stg_or_prd_env, is_test_env, is_hotseat_env,
+    FF_ENV_INDEXER, CGAP_ENV_INDEXER, is_indexer_env, indexer_env_for_env,
 )
-from dcicutils.misc_utils import PRINT, Retry
+from dcicutils.misc_utils import PRINT, Retry, apply_dict_overrides
+from dcicutils.qa_utils import override_environ
+
 
 # constants associated with EB-related APIs
 EB_CONFIGURATION_SETTINGS = 'ConfigurationSettings'
@@ -390,7 +394,7 @@ class EBDeployer:
             exit(cls.deploy_indexer(args.env, args.application_version))
 
 
-class Deployer:  # XXX: this should change. It is not a deployer, it is a configuration file generator. - Will
+class IniFileManager:
 
     TEMPLATE_DIR = None
     INI_FILE_NAME = "production.ini"
@@ -400,7 +404,7 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
     def build_ini_file_from_template(cls, template_file_name, init_file_name,
                                      bs_env=None, bs_mirror_env=None, s3_bucket_env=None,
                                      data_set=None, es_server=None, es_namespace=None,
-                                     indexer=None, index_server=None):
+                                     indexer=None, index_server=None, sentry_dsn=None):
         """
         Builds a .ini file from a given template file.
 
@@ -425,7 +429,8 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
                                                data_set=data_set,
                                                es_server=es_server,
                                                es_namespace=es_namespace,
-                                               indexer=indexer)
+                                               indexer=indexer,
+                                               sentry_dsn=sentry_dsn)
 
     # Ref: https://stackoverflow.com/questions/19911123/how-can-you-get-the-elastic-beanstalk-application-version-in-your-application  # noqa: E501
     EB_MANIFEST_FILENAME = "/opt/elasticbeanstalk/deploy/manifest"
@@ -470,7 +475,8 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
     @classmethod
     def build_ini_stream_from_template(cls, template_file_name, init_file_stream,
                                        bs_env=None, bs_mirror_env=None, s3_bucket_env=None, data_set=None,
-                                       es_server=None, es_namespace=None, indexer=None, index_server=None):
+                                       es_server=None, es_namespace=None, indexer=None, index_server=None,
+                                       sentry_dsn=None):
         """
         Sends output to init_file_stream corresponding to the data noe would want in an ini file
         for the given template_file_name and available environment variables.
@@ -485,7 +491,8 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
             es_server: The name of an es server to use.
             es_namespace: The namespace to use on the es server. If None, this uses the bs_env.
             indexer: Whether or not we are building an ini file for an indexer.
-            indexer: Whether or not we are building an ini file for an index server.
+            index_server: Whether or not we are building an ini file for an index server.
+            sentry_dsn: A sentry DSN specifier, or the empty string if none is desired.
 
         Returns: None
 
@@ -498,6 +505,7 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
         data_set = data_set or os.environ.get("ENCODED_DATA_SET",
                                               data_set_for_env(bs_env) or "MISSING_ENCODED_DATA_SET")
         es_namespace = es_namespace or os.environ.get("ENCODED_ES_NAMESPACE", bs_env)
+        sentry_dsn = sentry_dsn or os.environ.get("ENCODED_SENTRY_DSN", "")
         # Set ENCODED_INDEXER to 'true' to deploy an indexer.
         # If the value is missing, the empty string, or any other thing besides 'true' (in any case),
         # this value will default to the empty string, causing the line not to appear in the output file
@@ -538,25 +546,24 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
             'ES_NAMESPACE': es_namespace,
             'INDEXER': indexer,
             'INDEX_SERVER': index_server,
+            'SENTRY_DSN': sentry_dsn,
         }
 
         # if we specify an indexer name for bs_env, we did the deployment wrong and should bail
         if bs_env in INDEXER_ENVS:
             raise RuntimeError("Deployed with bs_env %s, which is an indexer env."
-                               "Re-deploy with the env you want to index and set the 'ENCODED_INDEXER'"
-                               "environment variable." % bs_env)
+                               " Re-deploy with the env you want to index and set the 'ENCODED_INDEXER'"
+                               " environment variable." % bs_env)
 
         # We assume these variables are not set, but best to check first. Confusion might result otherwise.
-        for extra_var in extra_vars:
-            if extra_var in os.environ:
-                raise RuntimeError("The environment variable %s is already set to %s."
-                                   % (extra_var, os.environ[extra_var]))
+        for extra_var, extra_var_val in extra_vars.items():
+            if extra_var in os.environ and extra_var_val != os.environ[extra_var]:
+                raise RuntimeError("The environment variable %s is already set to %s,"
+                                   " but you are trying to set it to %s."
+                                   % (extra_var, os.environ[extra_var], extra_var_val))
 
-        try:
-
-            # When we've checked everything, go ahead and do the bindings.
-            for var, val in extra_vars.items():
-                os.environ[var] = val
+        # When we've checked everything, go ahead and do the bindings.
+        with override_environ(**extra_vars):
 
             with io.open(template_file_name, 'r') as template_fp:
                 for line in template_fp:
@@ -569,13 +576,6 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
                     if not cls.omittable(line, expanded_line):
                         init_file_stream.write(expanded_line)
 
-        finally:
-
-            for key in extra_vars.keys():
-                # Let's be tidy and put things back the way they were before.
-                # Most things probably don't care, but testing might.
-                del os.environ[key]
-
     @classmethod
     def any_environment_template_filename(cls):
         file = os.path.join(cls.TEMPLATE_DIR, "any.ini")
@@ -658,6 +658,9 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
                                 help="whether this is a standalone indexing server, only doing indexing",
                                 choices=["true", "false"],
                                 default=None)
+            parser.add_argument("--sentry_dsn",
+                                help="a sentry DSN",
+                                default=None)
             args = parser.parse_args()
             template_file_name = (cls.any_environment_template_filename()
                                   if args.use_any
@@ -669,11 +672,94 @@ class Deployer:  # XXX: this should change. It is not a deployer, it is a config
                                              bs_env=args.bs_env, bs_mirror_env=args.bs_mirror_env,
                                              s3_bucket_env=args.s3_bucket_env, data_set=args.data_set,
                                              es_server=args.es_server, es_namespace=args.es_namespace,
-                                             indexer=args.indexer, index_server=args.index_server)
+                                             indexer=args.indexer, index_server=args.index_server,
+                                             sentry_dsn=args.sentry_dsn)
         except Exception as e:
             PRINT("Error (%s): %s" % (e.__class__.__name__, e))
             sys.exit(1)
 
 
+# The name Deployer is deprecated. Please use IniFileManager instead of Deployer.
+Deployer = IniFileManager
+
+
+class DeploymentFailure(RuntimeError):
+    pass
+
+
+class CreateMappingOnDeployManager:
+
+    # Set SKIP to True to skip the create_mapping step.
+
+    DEFAULT_DEPLOYMENT_OPTIONS = {'SKIP': False, 'STRICT': False, 'WIPE_ES': False}
+    STAGING_DEPLOYMENT_OPTION_OVERRIDES = {'WIPE_ES': True, 'STRICT': True}
+    HOTSEAT_DEPLOYMENT_OPTION_OVERRIDES = {'SKIP': True, 'STRICT': True}
+    OTHER_TEST_DEPLOYMENT_OPTION_OVERRIDES = {'WIPE_ES': True}
+    OTHER_PROD_DEPLOYMENT_OPTION_OVERRIDES = {'SKIP': True}
+
+    @classmethod
+    def _summarize_deploy_options(cls, options):
+        return "SKIP" if options['SKIP'] else ",".join(k for k in ('STRICT', 'WIPE_ES') if options[k]) or "default"
+
+    @classmethod
+    def get_deploy_config(cls, *, env, args, log, client=None, allow_other_prod=False):
+        """
+        Returns a dictionary describing appropriate options for creating mapping on deploy of a non-prd server.
+            {)
+                "ENV_NAME": env,    # what environment we're working with
+                "WIPE_ES": <bool>,  # whether to wipe ElasticSearch before reindex
+                "STRICT": <bool>,   # whether to do a 'strict' reindex
+                "SKIP": <bool>,     # whether to skip this step (notwithstanding other options)
+            }
+        NOTE WELL: This method will fail (raising DeploymentFailure) if called on production.
+        """
+
+        deploy_cfg = {
+            'ENV_NAME': env
+        }
+
+        current_prod_env = compute_ff_prd_env() if is_fourfront_env(env) else compute_cgap_prd_env()
+
+        apply_dict_overrides(deploy_cfg, **cls.DEFAULT_DEPLOYMENT_OPTIONS)
+        apply_dict_overrides(deploy_cfg, WIPE_ES=args.wipe_es, SKIP=args.skip, STRICT=args.strict)
+
+        if env == get_standard_mirror_env(current_prod_env):
+            description = "currently the staging environment"
+            apply_dict_overrides(deploy_cfg, **cls.STAGING_DEPLOYMENT_OPTION_OVERRIDES)
+        elif is_stg_or_prd_env(env):
+            if env == current_prod_env:
+                log.info("Environment %s is currently the production environment."
+                         " Something is definitely wrong. We never deploy there, we always CNAME swap."
+                         " This deploy cannot proceed. DeploymentFailure will be raised." % env)
+                raise DeploymentFailure('Tried to run %s on production.' % client or cls.__name__)
+            elif allow_other_prod:
+                description = "an uncorrelated production-class environment (neither production nor its staging mirror)"
+                apply_dict_overrides(deploy_cfg, **cls.OTHER_PROD_DEPLOYMENT_OPTION_OVERRIDES)
+            else:
+                log.info("Environment %s is an uncorrelated production-class environment."
+                         " Something is definitely wrong."
+                         " This deploy cannot proceed. DeploymentFailure will be raised." % env)
+                raise DeploymentFailure("Tried to run %s on a production-class environment"
+                                        " (neither production nor its staging mirror)." % client or cls.__name__)
+        elif is_test_env(env):
+            if is_hotseat_env(env):
+                description = "a hotseat test environment"
+                apply_dict_overrides(deploy_cfg, **cls.HOTSEAT_DEPLOYMENT_OPTION_OVERRIDES)
+            else:  # webdev or mastertest
+                description = "a non-hotseat test environment"
+                apply_dict_overrides(deploy_cfg, **cls.OTHER_TEST_DEPLOYMENT_OPTION_OVERRIDES)
+        else:
+            description = "an unrecognized environment"
+        log.info('Environment %s is %s. Processing mode: %s'
+                 % (env, description, cls._summarize_deploy_options(deploy_cfg)))
+        return deploy_cfg
+
+    @staticmethod
+    def add_argparse_arguments(parser):
+        parser.add_argument('--wipe-es', help="Specify to wipe ES", action='store_true', default=None)
+        parser.add_argument('--skip', help='Specify to skip this step altogether', default=None)
+        parser.add_argument('--strict', help='Specify to do a strict reindex', default=False)
+
+
 if __name__ == "__main__":
     EBDeployer.main()  # noqa - this is just for debugging
