@@ -5,11 +5,14 @@ This file contains functions that might be generally useful.
 import contextlib
 import datetime
 import functools
+import inspect
 import math
 import io
 import os
 import logging
 import pytz
+import rfc3986.validators
+import rfc3986.exceptions
 import time
 import warnings
 import webtest  # importing the library makes it easier to mock testing
@@ -48,6 +51,44 @@ PRINT = _PRINT()
 PRINT.__name__ = 'PRINT'
 
 
+absolute_uri_validator = (
+    rfc3986.validators.Validator()
+    # Validation qualifiers
+    .allow_schemes('http', 'https')
+    # TODO: We might want to consider the possibility of forbidding the use of a password. -kmp 20-Apr-2021
+    # .forbid_use_of_password()
+    .require_presence_of('scheme', 'host')
+    .check_validity_of('scheme', 'host', 'path'))
+
+
+def is_valid_absolute_uri(text):
+    """
+    Returns True if the given text is a string in the proper format to be an 'absolute' URI,
+    by which we mean the URI has a scheme (http or https) and a host specification.
+
+    For more info, see "Uniform Resource Identifier (URI): Generic Syntax" at https://tools.ietf.org/html/rfc3986
+    """
+    # Technically something like 'foo/bar.html' is also a URI, but it is a relative one, and
+    # the intended use of this function is to verify the URI specification of a resource on the web,
+    # independent of browser context, so a relative specification would be meaningless. We can add
+    # a separate operation for that later if we need one.
+    #
+    # We don't use rfc3987 (IRIs) both because it allows some resource locators we're not sure we're
+    # committed to accepting. Wikipedia, in https://en.wikipedia.org/wiki/Internationalized_Resource_Identifier,
+    # hints that there is some controversy about whether IRIs are even a good idea. We can revisit the idea if
+    # someone is demanding it. (And, as a practical matter, the rfc3987 library has a problematic license.)
+    # -kmp 21-Apr-2021
+    try:
+        uri_ref = rfc3986.uri_reference(text)
+    except ValueError:
+        return False
+    try:
+        absolute_uri_validator.validate(uri_ref)
+        return True
+    except rfc3986.exceptions.ValidationError:
+        return False
+
+
 class VirtualAppError(Exception):
     """ Special Exception to be raised by VirtualApp that contains some additional info """
 
@@ -69,7 +110,23 @@ class VirtualAppError(Exception):
         return self.__repr__()
 
 
-class _VirtualAppHelper(webtest.TestApp):  # effectively disguises 'TestApp'
+class TestApp(webtest.TestApp):
+    """
+    Equivalent to webtest.TestApp, but pytest will not let the name confuse into thinking it's a test case.
+
+    A test case in PyTest is something that contains "test" in its name. We didn't pick the name TestApp,
+    but there may be tools that want to use TestApp for testing, and so this is a better place to inherit from,
+    since we've added an appropriate declaration to keep PyTest from confusing it with a TestCase.
+    """
+
+    __test__ = False  # This declaration asserts to PyTest that this is not a test case.
+
+
+class _VirtualAppHelper(TestApp):
+    """
+    A helper class equivalent to webtest.TestApp, except that it isn't intended for test use.
+    """
+
     pass
 
 
@@ -112,6 +169,20 @@ class VirtualApp:
         except webtest.AppError as e:
             raise VirtualAppError(msg='HTTP GET failed.', url=url, body='<empty>', raw_exception=e)
 
+    def post(self, url, obj, **kwargs):
+        """ Wrapper for TestApp.post that logs the outgoing POST
+
+        :param url: url to POST to
+        :param obj: object body to POST
+        :param kwargs: args to pass to the POST
+        :return: result of POST
+        """
+        logging.info('OUTGOING HTTP POST on url: %s with object: %s' % (url, obj))
+        try:
+            return self.wrapped_app.post(url, obj, **kwargs)
+        except webtest.AppError as e:
+            raise VirtualAppError(msg='HTTP POST failed.', url=url, body=obj, raw_exception=e)
+
     def post_json(self, url, obj, **kwargs):
         """ Wrapper for TestApp.post_json that logs the outgoing POST
 
@@ -164,6 +235,33 @@ class VirtualApp:
         return self.wrapped_app.app
 
 
+def exported(*variables):
+    """
+    This function does nothing but is used for declaration purposes.
+    It is useful for the situation where one module imports names from another module merely to allow
+    functions in another module to import them, usually for legacy compatibility.
+    Otherwise, the import might look unnecessary.
+    e.g.,
+
+    ---file1.py---
+    def identity(x):
+        return x
+
+    ---file2.py---
+    from .file1 import identity
+    from dcicutils.misc_utils import exported
+
+    # This function used to be defined here, but now is defined in file1.py
+    exported(identity)
+
+    ---file3.py---
+    # This file has not been updated to realize that file1.py is the new home of identity.
+    from .file2 import identity
+    print("one=", identity(1))
+    """
+    ignored(variables)
+
+
 def ignored(*args, **kwargs):
     """
     This is useful for defeating flake warnings.
@@ -182,6 +280,24 @@ def ignored(*args, **kwargs):
     return args, kwargs
 
 
+def ignorable(*args, **kwargs):
+    """
+    This is useful for defeating flake warnings.
+    Call this function to use values that really might be ignored.
+    This is intended as a declaration that variables are or might be intentionally ignored,
+    but no enforcement of that is done. Some sample uses:
+
+    def foo(x, y):
+        ignorable(x, y)  # so flake8 won't complain about unused vars, whether or not next line is commented out.
+        # print(x, y)
+        return 3
+
+    foo_synonym = foo
+    ignorable(foo_synonym)  # We might or might not use foo_synonym, but we don't want it reported as unused
+    """
+    return args, kwargs
+
+
 def get_setting_from_context(settings, ini_var, env_var=None, default=None):
     """
     This gets a value from either an environment variable or a config file.
@@ -314,7 +430,7 @@ class Retry:
             wait_multiplier: A multiplier by which the number of wait_seconds is adjusted on each retry.
         """
 
-        def decorator(function):
+        def _decorator(function):
             function_name = name_key or function.__name__
             function_profile = cls.RetryOptions(
                 retries_allowed=cls._defaulted(retries_allowed, cls.DEFAULT_RETRIES_ALLOWED),
@@ -352,7 +468,7 @@ class Retry:
 
             return wrapped_function
 
-        return decorator
+        return _decorator
 
     @classmethod
     def retrying(cls, fn, retries_allowed=None, wait_seconds=None, wait_increment=None, wait_multiplier=None):
@@ -769,6 +885,48 @@ def environ_bool(var, default=False):
         return os.environ[var].lower() == "true"
 
 
+@contextlib.contextmanager
+def override_environ(**overrides):
+    """
+    Overrides os.environ for the dynamic extent of the call, using the specified values.
+    A value of None means to delete the property temporarily.
+    (This uses override_dict to do the actual overriding. See notes for that function about lack of thread safety.)
+    """
+    with override_dict(os.environ, **overrides):
+        yield
+
+
+@contextlib.contextmanager
+def override_dict(d, **overrides):
+    """
+    Overrides the given dictionary for the dynamic extent of the call, using the specified values.
+    A value of None means to delete the property temporarily.
+
+    This function is not threadsafe because it dynamically assigns and de-assigns parts of a dictionary.
+    It should be reserved for use in test functions or command line tools or other contexts that are known
+    to be single-threaded, or at least not competing for the resource of the dictionary. (It would be threadsafe
+    to use a dictionary that is only owned by the current process.)
+    """
+    to_delete = []
+    to_restore = {}
+    try:
+        for k, v in overrides.items():
+            if k in d:
+                to_restore[k] = d[k]
+            else:
+                to_delete.append(k)
+            if v is None:
+                d.pop(k, None)  # Delete key k, tolerating it being already gone
+            else:
+                d[k] = v
+        yield
+    finally:
+        for k in to_delete:
+            d.pop(k, None)  # Delete key k, tolerating it being already gone
+        for k, v in to_restore.items():
+            d[k] = v
+
+
 def check_true(test_value, message, error_class=None):
     """
     If the first argument does not evaluate to a true value, an error is raised.
@@ -785,6 +943,27 @@ def check_true(test_value, message, error_class=None):
         raise error_class(message)
 
 
+def remove_element(elem, lst, raise_error=True):
+    """
+    Returns a shallow copy of the given list with the first occurrence of the given element removed.
+
+    If the element doesn't occur in the list, an error is raised unless given raise_error=False,
+    in which case a shallow copy of the original list is returned (with no elements removed).
+
+    :param elem: an object
+    :param lst: a list
+    :param raise_error: a boolean (default True)
+    """
+
+    result = lst.copy()
+    try:
+        result.remove(elem)
+    except ValueError:
+        if raise_error:
+            raise
+    return result
+
+
 def remove_prefix(prefix, text, required=False):
     if not text.startswith(prefix):
         if required:
@@ -821,6 +1000,21 @@ def obsolete(func, fail=True):
     return inner
 
 
+def ancestor_classes(cls, reverse=False):
+    result = list(cls.__mro__[1:])
+    if reverse:
+        result.reverse()
+    return result
+
+
+def is_proper_subclass(cls, maybe_proper_superclass):
+    """
+    Returns true of its first argument is a subclass of the second argument, but is not that class itself.
+    (Every class is a subclass of itself, but no class is a 'proper subclass' of itself.)
+    """
+    return cls is not maybe_proper_superclass and issubclass(cls, maybe_proper_superclass)
+
+
 def full_class_name(obj):
     """
     Returns the fully-qualified name of the class of the given obj (an object).
@@ -859,6 +1053,44 @@ def constantly(value):
     return fn
 
 
+def identity(x):
+    """Returns its argument."""
+    return x
+
+
+def count_if(filter, seq):  # noQA - that's right, we're shadowing the built-in Python function 'filter'.
+    return sum(1 for x in seq if filter(x))
+
+
+def count(seq, filter=None):  # noQA - that's right, we're shadowing the built-in Python function 'filter'.
+    return count_if(filter or identity, seq)
+
+
+def find_associations(data, **kwargs):
+    found = []
+    for datum in data:
+        mismatch = False
+        for k, v in kwargs.items():
+            defaulted_val = datum.get(k)
+            if not (v(defaulted_val) if callable(v) else (v == defaulted_val)):
+                mismatch = True
+                break
+        if not mismatch:
+            found.append(datum)
+    return found
+
+
+def find_association(data, **kwargs):
+    results = find_associations(data, **kwargs)
+    n = len(results)
+    if n == 0:
+        return None
+    elif n == 1:
+        return results[0]
+    else:
+        raise ValueError("Got %s results when 1 was expected." % n)
+
+
 def keyword_as_title(keyword):
     """
     Given a dictionary key or other token-like keyword, return a prettier form of it use as a display title.
@@ -1078,6 +1310,77 @@ def url_path_join(*fragments):
     return result
 
 
+def _is_function_of_exactly_one_required_arg(x):
+    if not callable(x):
+        return False
+    argspec = inspect.getfullargspec(x)
+    return len(argspec.args) == 1 and not argspec.varargs and not argspec.defaults and not argspec.kwonlyargs
+
+
+def _apply_decorator(fn, *args, **kwargs):
+    """
+    This implements a fix to the decorator syntax where it gets fussy about whether @foo and @foo() are synonyms.
+    The price to be paid is you can't use it for decorators that take positional arguments.
+    """
+    if args and (kwargs or len(args) > 1):
+        # If both args and kwargs are in play, they have to have been passed explicitly like @foo(a1, k2=v2).
+        # If more than one positional is given, that has to be something like @foo(a1, a2, ...)
+        # Decorators using this function need to agree to only accept keyword arguments, so those cases can't happen.
+        # They can do this by using an optional first positional argument, as in 'def foo(x=3):',
+        # or they can do it by using a * as in 'def foo(*, x)' or if no arguments are desired, obviously, 'def foo():'.
+        raise SyntaxError("Positional arguments to decorator (@%s) not allowed here." % fn.__name__)
+    elif args:
+        arg0 = args[0]  # At this point, we know there is a single positional argument.
+        #
+        # Here there are two cases.
+        #
+        # (a) The user may have done @foo, in which case we will have a fn which is the value of foo,
+        #     but not the result of applying it.
+        #
+        # (b) Otherwise, the user has done @foo(), in which case what we'll have the function of one
+        #     argument that does the wrapping of the subsequent function or class.
+        #
+        # So since case (a) expects fn to be a function that tolerates zero arguments
+        # while case (b) expects fn to be a function that rejects positional arguments,
+        # we can call fn with the positional argument, arg0. If that argument is rejected with a TypeError,
+        # we know that it's really case (a) and that we need to call fn once with no arguments
+        # before retrying on arg0.
+        if _is_function_of_exactly_one_required_arg(fn):
+            # We are ready to wrap the function or class in arg0
+            return fn(arg0)
+        else:
+            # We are ALMOST ready to wrap the function or class in arg0,
+            # but first we have to call ourselves with no arguments as in case (a) described above.
+            return fn()(arg0)
+    else:
+        # Here we have kwargs = {...} from @foo(x=3, y=4, ...) or maybe no kwargs either @foo().
+        # Either way, we've already evaluated the foo(...) call, so all that remains is to call on our kwargs.
+        # (There are no args to call it on because we tested that above.)
+        return fn(**kwargs)
+
+
+def _decorator(decorator_function):
+    """See documentation for decorator."""
+    @functools.wraps(decorator_function)
+    def _wrap_decorator(*args, **kwargs):
+        return _apply_decorator(decorator_function, *args, **kwargs)
+    return _wrap_decorator
+
+
+@_decorator
+def decorator():
+    """
+    This defines a decorator, such that is can be used as either @foo or @foo()
+    PROVIDED THAT the function doing the decorating is not a function a single required argument,
+    since that would create an ambiguity that would inhibit the auto-correction this will do.
+
+    @decorator
+    def foo(...):
+        ...
+    """
+    return _decorator
+
+
 # Deprecated names, still supported for a while.
 HMS_TZ = REF_TZ
 hms_now = ref_now
